defstar {
	name {adint}
	domain {SDF}
        desc {
A/D integrator:
This star integrates the difference of the input signal U
and the intervallwise constant feedback signal fb with a
high order implicit Adams-Moulton method. See the manual
for further details.
        }
        version {@(#)SDFadint.pl	1.0 AUG/6/98}
	author { F. Schilder }
	copyright {
Copyright (c) 1998 by Frank Schilder. All rights reserved.
See the file $PTOLEMY/copyright for copyright notice,
limitation of liability, and disclaimer of warranty provisions.
	}
	location { A/D converter simulation package }

        hinclude  { <Matrix.h> }
        ccinclude { <math.h> }
        input {
            name { uInput }
            type { float_matrix_env }
        }
        input {
            name { feedback }
            type { float }
        }

        output {
            name { uOutput }
            type { float_matrix_env }
        }

        protected {
            double f[6], U, U_in;
            FloatMatrix lastAdintIn;
            FloatMatrix &adintOut;
            Envelope outMatrixEnv;
            int State, N;
        }
        state {
            name { U0 }
            default { 0.0 }
            type { float }
            desc { initial value U(0) }
            attributes { A_CONSTANT | A_SETTABLE }
        }
        state {
            name { T }
            default { 1.0 }
            type { float }
            desc { (inverse) integration factor }
            attributes { A_CONSTANT | A_SETTABLE }
        }
        state {
            name { dt }
            default { 0.0 }
            type { float }
            desc { time-step-size }
            attributes { A_CONSTANT | A_SETTABLE }
        }

        setup {
            State = 5;
            if(dt <= 0.0) Error::abortRun(*this, "dt not initialized correctly");
        }
        go {
            int i,j;
            Envelope inMatrixEnv;
            (uInput%0).getMessage(inMatrixEnv);
            const FloatMatrix& adintIn = *(const FloatMatrix *) inMatrixEnv.myData();

            // init: create the matrices after the first particle has been read
            // so we know our stage and thus the dimension of the output array
            if(State == 5) {
                N=adintIn.numCols();
                lastAdintIn=FloatMatrix(N+1, N+1);
                for(i=0; i<=N; i++)
                  for(j=0; j<=N; j++)
                    lastAdintIn[i][j] = 0.0;
                adintOut = lastAdintIn;
            }

            // 1) read the input
            U_in           = adintIn[0][(N==1) ? 0 : 1]/double(T);
            adintOut[1][0] = double(feedback%0);
            for(i=1; i<N; i++)
              for(j=0; j<=i; j++)
                adintOut[i+1][j] = adintIn[i][j];

            // 2) transform the input
            for(i=1; i<=N; i++)
              for(j=0; j<=N-i; j++)
                adintOut[i+j][j] *= double(dt)/(double(i)*double(T));

            // 3) calculate nth integral over U
            switch( State )
            {
                case 0: { Adams6(); break;}
                case 1: { Adams5(); break;}
                case 2: { Adams4(); break;}
                case 3: { Adams3(); break;}
                case 4: { Adams2(); break;}
                case 5: { Adams1(); break;}
            }
            adintOut[0][1] = U;

            // 4+5) calculate the multiple integrals and the output signal
            // the multiple integrals are insertet in calculateFeedback() (see below)
            adintOut[0][0] = U-calculateFeedback();

            // 6) save one copy for the next iteration
            lastAdintIn = adintOut;

            // 7) fire the output through the output port
            uOutput%0 << adintOut;
        }

        conscalls { adintOut(*new FloatMatrix), outMatrixEnv(adintOut) }
        method {
            name { Adams6 }
            code {
                memmove(f+1, f, sizeof(double)*5);
                f[0] = U_in;
                U += double(dt)/1440.0 * ( 475.0*f[0] + 1427.0*f[1] - 798.0*f[2]
                                         + 482.0*f[3] -  173.0*f[4] +  27.0*f[5]);
            }
        }
        method {
            name { Adams5 }
            code {
                memmove(f+1, f, sizeof(double)*4);
                f[0] = U_in;
                U += double(dt)/720.0 * ( 251.0*f[0] + 646.0*f[1] - 264.0*f[2]
                                                     + 106.0*f[3] -  19.0*f[4]);
                State--;
            }
        }
        method {
            name { Adams4 }
            code {
                memmove(f+1, f, sizeof(double)*3);
                f[0] = U_in;
                U += double(dt)/24.0 * (9.0*f[0] + 19.0*f[1] - 5.0*f[2] + f[3]);
                State--;
            }
        }
        method {
            name { Adams3 }
            code {
                memmove(f+1, f, sizeof(double)*2);
                f[0] = U_in;
                U += double(dt)/12.0 * (5.0*f[0] + 8.0*f[1] - f[2]);
                State--;
            }
        }
        method {
            name { Adams2 }
            code {
                f[1] = f[0];
                f[0] = U_in;
                U += 0.5 * double(dt) * (f[0] + f[1]);
                State--;
            }
        }
        method {
            name { Adams1 }
            code {
                f[0] = U_in;
                U = U0;
                State--;
            }
        }
        method {
            name { calculateFeedback }
            type { double }
            code {
                double res = 0.0;
                int i;
                
                // calculate the feedback required for calculating the output signal
                // (step 5, see above)
                for(i=1; i<=N; i++) res += ( adintOut[i][i]=mInt(i, lastAdintIn[i]) );
                return res;
            }
        }
        method {
            name { mInt }
            arglist { "(const int n, const double* oldIs)" }
            type { double }
            code {
                int i;
                double res=0.0;
                
                // 4) calculate the multiple integrals over the feedbacks
                for(i=0; i<=n; i++) res += oldIs[i];
                return res;
            }
        }
}
